// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_comm_ServerSocket_hpp
#define anna_comm_ServerSocket_hpp

#include <anna/core/util/Recycler.hpp>
#include <anna/core/util/Millisecond.hpp>

#include <anna/comm/Socket.hpp>
#include <anna/comm/internal/LocalConnection.hpp>

namespace anna {

namespace comm {

class INetAddress;
class BinderSocket;

/**
   Esta clase implementa el servidor de Socket. Un servidor de Socket espera las peticiones de entrada procedentes
   de la red y realiza las operaciones necesarias para tratar la operacion y posiblemente devolver un resultado
   al peticionario.

   El 'backlog' de un servidor de socket define la longitud maxima que la cola de mensajes pendientes puede alcanzar.
   Si cliente realiza una peticion mediante una conexion UF_UNIX cuando la cola de mensajes del servidor esta llena
   recibira un error ECONNREFUSED, aunque normalmente el protocolo subyacente se encarga de retransmitir la peticion
   un numero indeterminado de veces.

   @see Communicator::accept
*/
class ServerSocket : public Socket {
public:
  /**
     numero de milisegundos por defecto que espera antes de dar por fallida una asociacion a una
     direccion.
  */
  static const Millisecond DefaultBindDelay;

  typedef Recycler<LocalConnection>::iterator iterator;

  /**
     Tamao de la cola de mensajes tomado por defecto.
  */
  static const int defaultBacklog = 30;

  /**
     Crea un servidor de socket liberado.
     \param transportFactory Factoria de protocolos de transporte a usar por este sockets.
     \param domain Dominio del socket.
     \param type Tipo de socket.
     \warning La Factoria de protocolos debe estar disponible mientras el Socket esta activo.
  */
  ServerSocket(TransportFactory* transportFactory = NULL, Domain::_v domain = Socket::Domain::Inet, Type::_v type = Socket::Type::Stream) :
    Socket(domain, type, transportFactory),
    a_backlog(defaultBacklog),
    a_sharedBind(false),
    a_binderSocket(NULL),
    a_msBindDelay(DefaultBindDelay) {}

  /**
     Crea un servidor de socket conectado a la direccion y puerto indicado y con la longitud de cola maxima indicada.

     \param localAddress Direccion Local en la que atendera las peticiones.
     \param transportFactory Factoria de protocolos de transporte a usar por este sockets.
     \param sharedBind \em true Si puede haber mas de una aplicacion escuchando es esta misma direccion.
     \warning La factoria de protocolos debe estar disponible mientras el Socket esta activo.
  */
  ServerSocket(const INetAddress& localAddress, const bool sharedBind, TransportFactory* transportFactory = NULL) :
    Socket(localAddress, Socket::Type::Stream, transportFactory),
    a_backlog(defaultBacklog),
    a_sharedBind(sharedBind),
    a_binderSocket(NULL),
    a_msBindDelay(DefaultBindDelay) {}

  /**
     Destructor.
  */
  virtual ~ServerSocket();

  /**
     Devuelve el numero de milisegundos esperado al hacer el bind compartido con este sockets
     antes de considerar que no se puede atender peticiones en esa direccion.
     \return el numero de milisegundos esperado al hacer el bind compartido con este sockets
     antes de considerar que no se puede atender peticiones en esa direccion.
  */
  const Millisecond &getBindDelay() throw() { return a_msBindDelay; }

  /**
     Devuelve el socket asociado a este socket de bind compartido.
     \return El socket asociado a este socket de bind compartido. Puede ser NULL.
  */
  BinderSocket* getBinderSocket() throw() { return a_binderSocket; }

  /**
     Devuelve el modo de asociacion de este socket.
     \return El modo de asociacion de este socket.
     \warning El valor devuelto por este metodo solo sera valido despues de ejecutar correctamente
     el metodo #bind
  */
  bool isSharedBind() const throw() { return a_sharedBind; }

  /**
     Establece el tamao de la cola de mensajes.
     \param backlog Tamao de la cola de mensajes.
  */
  void setBacklog(const int backlog) throw() { a_backlog = backlog; }

  /**
     Establece el numero de milisegundos esperado al hacer el bind con este sockets antes
     de considerar que no se puede realizar la conexion.

     \param bindDelay numero de milisegundos esperado al hacer la conexion con este
     sockets antes de considerar que no se puede realizar la conexion.

     \see #DefaultBindDelay

     \warning Solo tendra efecto en el caso de que el socket tenga activado el modo de 'Bind compartido'.
  */
  void setBindDelay(const Millisecond &bindDelay) throw() { a_msBindDelay = bindDelay; }

  /**
     Metodo que debemos invocar una vez que el ServerSocket esta establecido.
     \warning Normalmente se invocar�desde Communicator::attach(ServerSocket&,bool)
  */
  virtual void prepare() throw(RuntimeException);

  /**
     Comprueba la conexion establecida y acepta las peticiones. Esta funcin puede bloquear al thread
     que la invoca mientras no llegue una peticion de apertura de conexion si no hemos establecido el modo
     no-bloqueante del socket. Ver Socket::setBlockingMode.

     Cada uno de los socket obtenidos con este Metodo debe ser liberado con #release cuando ya no sean
     necesarios.

     @return Nueva conexion aceptada, pendiente de pasar los filtros de acceptacion.

     \warning Exclusivamente uso interno.
  */
  LocalConnection* accept() throw(RuntimeException);

  /**
     Libera los recursos del socket recibido como parametro.

     @param localConnection Socket del que vamos a liberar los recursos. Esta instancia deberia haberse obtenido
     mediante el Metodo #accept, ya que en otro caso los resultados no estan definidos.

     \warning Exclusivamente uso interno.
  */
  void release(LocalConnection* localConnection) throw(RuntimeException);

  /**
     Devuelve un iterador al primer LocalConnection definido.
     \return un iterador al primer LocalConnection definido.
  */
  iterator begin() throw() { return a_localConnections.begin(); }

  /**
     Devuelve un iterador al ultimo LocalConnection definido.
     \return un iterador al ultimo LocalConnection definido.
  */
  iterator end() throw() { return a_localConnections.end(); }

  /**
     Devuelve una cadena con la informacion referente a este socket.
     @return Una cadena con la informacion referente a este socket.
  */
  virtual std::string asString() const throw();

  /**
     Devuelve un nodo XML con la informacion referente a este objeto.
     \param parent Nodo XML a partir del cual introducir la informacion.
     \return Un nodo XML con la informacion referente a este objeto.
  */
  virtual xml::Node* asXML(xml::Node* parent) const throw(RuntimeException);

  /**
     Metodo manejador de evento que permite ajustar el funcionamiento cuando el nucleo de
     anna.comm notifica que ha detectado una peticion de conexion desde un proceso
     remoto a un ServerSocket asociado al comunicador. Permite independencia de dicho
     comunicador y complementa el control de aceptacion (por defecto, se devuelve true
     para no influir en la implementacion del metodo analogo en el comunicador).

     @param clientSocket Socket cliente que solicita acceso.

     \return \em true si la conexion es acceptada a \em false en otro caso, en cuyo caso se liberaran
     automaticamente todos los recursos asociados a la peticion de conexion.
  */
  virtual bool eventAcceptConnection(const ClientSocket &clientSocket) throw(RuntimeException) { return true; }

//   /**
//       Informa sobre la rotura de una conexion que se creo a partir de un ServerSocket
//
//      @param localConnection Socket que se acepto sobre el server socket y que se ha roto.
//   */
//   virtual void eventBreakLocalConnection (LocalConnection* localConnection) throw (RuntimeException) {;}


  /**
     Devuelve una referencia al contenido apuntado por el iterador.
     \return una referencia al contenido apuntado por el iterador.
  */
  static LocalConnection* localConnection(iterator& ii) throw() { return Recycler<LocalConnection>::data(ii); }

  /**
     Devuelve el nombre logico de esta clase.
     \return el nombre logico de esta clase.
  */
  static const char* className() throw() { return "anna::comm::ServerSocket"; }

protected:
  /**
     Crea un servidor de socket conectado a la direccion y puerto indicado y con la longitud de cola maxima indicada.

     \param transportFactory Factoria de protocolos de transporte a usar por este sockets.
     \param localAddress Puede ser usado para limitar la direccion por la que atendiende peticiones un servidor de socket
     instalado en una maquina con mas de una direccion.
     \param sharedBind \em true Si puede haber mas de una aplicacion escuchando es esta misma direccion.
     \param type Tipo de socket.
     \warning La Factoria de protocolos debe estar disponible mientras el Socket esta activo.
  */
  ServerSocket(const INetAddress& localAddress, const bool sharedBind, Type::_v type, TransportFactory* transportFactory = NULL) :
    Socket(localAddress, type, transportFactory),
    a_backlog(defaultBacklog),
    a_sharedBind(false),
    a_binderSocket(NULL),
    a_msBindDelay(DefaultBindDelay) {}

private:
  int a_backlog;
  Recycler <LocalConnection> a_localConnections;
  const bool a_sharedBind;
  BinderSocket* a_binderSocket;
  Millisecond a_msBindDelay;

  virtual ClientSocket* allocateClientSocket() const throw();
  int do_bind(const struct sockaddr *, const int len) throw(RuntimeException);

  friend class BinderSocket;
};

}
}


#endif



