// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_core_define_autoenum_hpp
#define anna_core_define_autoenum_hpp

#include <string.h>

#include <anna/core/RuntimeException.hpp>

/**
 * \page declare_enum
 *
 * Facilita la declaracin de enumerados y sus correspondientes mtodos de conversin.
 * Los valores de los enumerados debe incluir el valor \em None, que ser usado en el mtodo
 * asEnum (const char*) incorporado por esta macro en caso de no encontrar ninguna correspondencia.
 *
 * El numerado desde el que se invoque esta macro incluir de forma automtica los siguiente mtodos:
 *
 * \li _v Transport::asEnum (const char* str): que compara la cadena recibida con las cadenas establecidas al asignar los valores
 * literales del enumerado y devuelve su valor enumerado. Si no hay coincidencia devolver None.
 * \li _v Transport::asEnum (const std::string& str): que compara la cadena recibida con las cadenas establecidas al asignar los valores
 * literales del enumerado y devuelve su valor enumerado. Si no hay coincidencia devolver None.
 * \li _v Transport::asEnumEx (const char* str): que compara la cadena recibida con las cadenas establecidas al asignar los valores
 * literales del enumerado y devuelve su valor enumerado. Si no hay coincidencia lanzar una excepcin.
 * \li _v Transport::asEnumEx (const std::string& str): que compara la cadena recibida con las cadenas establecidas al asignar los valores
 * literales del enumerado y devuelve su valor enumerado. Si no hay coincidencia lanzar una excepcin.
 * \li const char* asCString (const _v) que devuelve el literal asociado al valor del enumerado recibido como parmetro. Puede devolver NULL.
 * \li const char* asNotNullCString (const _v) que devuelve el literal asociado al valor del enumerado recibido como parmetro. Puede devolver "<none>".
 * \li std::string asList () que devuelve el literal que la lista de valores vlidos.
 *
 * As quedara la declaracin del enumerado, que normalmente estar en un .h
 * \code
 * struct Transport {
 *    enum _v { None = -1, TCP, SCTP, UDP };
 *     anna_declare_enum (Transport);
 * };
 * \endcode
 *
 * \ref assign_enum
 * \ref item_enum
 */
#define anna_declare_enum(name) \
   static const char* literal []; \
   static int calculateSize () throw () { \
      int ii = 0; \
      while (literal [ii] != NULL) ii ++; \
      return ii; \
   } \
   static _v asEnum (const char* str) throw () { \
      for (int ii = 0; literal [ii] != NULL; ii ++) { \
         if (strcasecmp (str, literal [ii]) == 0) \
            return (_v) ii; \
      } \
      return None; \
   } \
   static _v asEnumEx (const char* str) throw (anna::RuntimeException) { \
      if (str == NULL) { \
         std::string msg (#name); \
         msg += "::asEnumEx | str can not be null"; \
         throw anna::RuntimeException (msg, __FILE__,__LINE__); \
      } \
      _v result = asEnum (str); \
      if (result == None) { \
         std::string msg (#name); \
         msg += " | Value: '"; \
         msg += str; \
         msg += "' is not valid | Valid values: "; \
         msg += asList (); \
         throw anna::RuntimeException (msg, __FILE__, __LINE__); \
      } \
      return result; \
   } \
   static _v asEnum (const std::string& str) throw () { return asEnum (str.c_str ()); } \
   static _v asEnumEx (const std::string& str) throw (anna::RuntimeException) { return asEnumEx (str.c_str ()); } \
   static const char* asCString (const _v v) throw () { return (v != None) ? literal [v]: NULL; } \
   static const char* asNotNullCString (const _v v) throw () { return (v != None) ? literal [v]: "<none>"; } \
   static std::string asList () throw () {\
      std::string result;\
      for (int ii = 0; literal [ii] != NULL; ii ++) { \
         if (ii == 0 && strcmp (literal [ii], "None") == 0) continue; \
         if (ii > 1) result += ' '; \
         result += "'"; result += literal [ii]; result += "'"; \
      } \
      return result; \
   }

/**
 * \page assign_enum
 *
 * Establece los literales asociados a cada uno de los valores de la enumeracin.
 *
 * Hay que completar el enumerado asignando los literales asociados a cada uno de los valores:
 *
 * \code
 * anna_assign_enum (diameter::avp::DiameterURI::Transport) = { "tcp", "sctp", "udp", NULL };
 * \endcode
 *
 * \ref declare_enum
 * \warning Debe incluir el valor \em NULL para indicar el fin de la lista de valores.
 */
#define anna_assign_enum(name) const char* name::literal []

/**
 * \page item_enum
 * Accede al literal definido por el Enumado y la posicion indicada como parmetro
 */
#define anna_item_enum(name,ii) name::literal[ii]


#endif
