// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#include <ctype.h>

#include <iostream>

#include <stdarg.h>

#include <libxml/parser.h>

#include <anna/core/functions.hpp>
#include <anna/core/tracing/Logger.hpp>
#include <anna/core/tracing/TraceMethod.hpp>
#include <anna/core/tracing/TraceFunction.hpp>
#include <anna/core/mt/Guard.hpp>

#include <anna/xml/xml.hpp>
#include <anna/xml/internal/sccs.hpp>

using namespace std;
using namespace anna;

char Parser::st_text [1024] = "";
RuntimeException* Parser::st_exception = NULL;

Parser::Parser() :
  a_root(NULL),
  a_encoding(false),
  a_buffEncode(NULL) {
  sccs::activate();
  a_context = (xmlValidCtxtPtr) xmlMalloc(sizeof(xmlValidCtxt));
  a_context->error = errorHandler;
  a_context->warning = warningHandler;
}

Parser::~Parser() {
  reset();
  xmlFree(a_context);
  delete a_buffEncode;
  delete a_root;
}

const Node* Parser::apply(const Document& document)
throw(RuntimeException) {
  LOGMETHOD(TraceMethod tf("anna::xml::Parser", "apply(Document)", ANNA_FILE_LOCATION));
  Guard guard(*this);
  reset();
  apply(const_cast <_xmlDoc*>(document.a_handle));
  return a_root;
}

const Node* Parser::apply(const Document& document, const DTD& dtd)
throw(RuntimeException) {
  LOGMETHOD(TraceMethod tf("anna::xml::Parser", "apply(Document,DTD)", ANNA_FILE_LOCATION));
  Guard guard(*this);
  reset();
  dtd.validate(a_context, const_cast <_xmlDoc*>(document.a_handle));
  apply(const_cast <_xmlDoc*>(document.a_handle));
  return a_root;
}

void Parser::apply(_xmlDoc* document)
throw(RuntimeException) {
  LOGMETHOD(TraceMethod tf("anna::xml::Parser", "apply (_xmlDoc)", ANNA_FILE_LOCATION));
  xmlNode* root;

  if((root = xmlDocGetRootElement(document)) == NULL)
    throw RuntimeException("Error interpreting XML document", ANNA_FILE_LOCATION);

  setupEncoding(document);

  if(a_root == NULL)
    a_root = new Node((const char*) root->name);
  else
    a_root->setName((const char*) root->name);

  for(xmlNs* ns = root->nsDef; ns != NULL; ns = ns->next)
    a_root->createNamespace((const char*) ns->prefix, (const char*) ns->href);

  if(root->ns)
    a_root->setNamespace(a_root->namespace_find((const char*) root->ns->prefix));

  attributes(a_root, root->properties);
  children(a_root, root->children);
}

void Parser::children(Node* node, xmlNode* xmlNode)
throw(RuntimeException) {
  Node* child;
  bool isSeparator;
  const char* w;

  while(xmlNode != NULL) {

    if((xmlNode->type) == XML_ELEMENT_NODE) {
      child  = node->createChild((const char*) xmlNode->name);

      for(xmlNs* ns = xmlNode->nsDef; ns != NULL; ns = ns->next)
        if(ns->prefix != NULL)
          a_root->createNamespace((const char*) ns->prefix, (const char*) ns->href);

      if(xmlNode->ns != NULL && xmlNode->ns->prefix != NULL)
        child->setNamespace(node->namespace_find((const char*) xmlNode->ns->prefix));

      attributes(child, xmlNode->properties);
      children(child, xmlNode->children);
    }
    else if((xmlNode->type) == XML_TEXT_NODE) {
      w = (const char*) xmlNode->content;
      isSeparator = true;

      while(*w != 0) {
        if(isspace(*w) == false) {
          isSeparator = false;
          break;
        }
        w ++;
      }

      if(isSeparator == false)
        node->createText(decode(xmlNode->content));
    }


    xmlNode = xmlNode->next;
  }
}

void Parser::attributes(Node* node, xmlAttr* attribute)
throw(RuntimeException) {
  const char* value;
  const Namespace* ns = NULL;

  while(attribute != NULL) {
    ns = (attribute->ns == NULL) ? NULL :  node->namespace_find((const char*) attribute->ns->prefix);
    value = (attribute->children != NULL) ? decode(attribute->children->content) : "";
    node->createAttribute((const char*) attribute->name, value, ns);
    attribute = attribute->next;
  }
}

/* static */
void Parser::errorHandler(void *ctx,  const char *msg, ...)
throw() {
  va_list ap;
  va_start(ap, msg);
  vsprintf(st_text, msg, ap);

  for(char* aux = st_text; *aux; aux ++)
    if(*aux == '\n')
      *aux = ' ';

  va_end(ap);
  Logger::write(Logger::Error, st_text, ANNA_FILE_LOCATION);
}

/* static */
void Parser::warningHandler(void *ctx,  const char *msg, ...)
throw() {
  va_list ap;

  if(Logger::isActive(Logger::Warning)) {
    va_start(ap, msg);
    vsprintf(st_text, msg, ap);

    for(char* aux = st_text; *aux; aux ++)
      if(*aux == '\n')
        *aux = ' ';

    va_end(ap);
    Logger::write(Logger::Warning, st_text, ANNA_FILE_LOCATION);
  }
}

// La metodo que invoque a este debe hacerlo con una secci�n cr�tica activa.
void Parser::reset()
throw() {
  if(a_root != NULL)
    a_root->clear();
}

void Parser::setupEncoding(_xmlDoc* document)
throw() {
  if(document->encoding != NULL) {
    a_encoding = true;

    if(a_buffEncode == NULL)
      a_buffEncode = new DataBlock(true);
    else
      a_buffEncode->clear();
  } else
    a_encoding = false;

  LOGDEBUG(
    string msg("anna::xml::Parser::setupEncoding | Encoding: ");
    msg += (document->encoding == NULL) ? "none" : (const char*) document->encoding;
    Logger::debug(msg, ANNA_FILE_LOCATION);
  );
}

const char* Parser::decode(const unsigned char* source)
throw() {
  if(a_encoding == false)
    return (const char*) source;

  if(xmlCheckUTF8(source) == 0)
    return (const char*) source;

  int srcLen = xmlStrlen(source);
  a_buffEncode->clear();
  a_buffEncode->allocate(srcLen + 1);
  unsigned char* target = (unsigned char*) a_buffEncode->getData();
  int trgLen = srcLen;
  UTF8Toisolat1(target, &trgLen, source, &srcLen);
  target [trgLen] = 0;
  return (const char*) target;
}
